from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Sequence

from secfetch.exceptions import SecFetchError


class FormTypeValidationError(SecFetchError):
    pass


@dataclass(frozen=True)
class FormTypesConfig:
    accepted_form_types: List[str]


def _load_packaged_form_types() -> FormTypesConfig:
    # Standard-library resource loading (no runtime deps).
    from importlib.resources import files

    raw = (files("secfetch.resources") / "form_types.json").read_text(encoding="utf-8")
    payload = json.loads(raw)
    accepted = payload.get("accepted_form_types")
    if not isinstance(accepted, list) or not all(isinstance(x, str) for x in accepted):
        raise FormTypeValidationError("Invalid packaged form_types.json format")
    return FormTypesConfig(accepted_form_types=sorted(set(accepted)))


def ensure_form_types_json(
    *, data_dir: Path, target_rel_path: Path = Path("config/form_types.json")
) -> Path:
    """
    Ensure a user-editable form types JSON exists under the data directory.
    Returns its path.
    """
    out_path = data_dir / target_rel_path
    if out_path.exists():
        return out_path

    cfg = _load_packaged_form_types()
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(
        json.dumps({"accepted_form_types": cfg.accepted_form_types}, indent=2, sort_keys=True) + "\n",
        encoding="utf-8",
    )
    return out_path


def load_accepted_form_types(*, data_dir: Optional[Path] = None) -> List[str]:
    """
    Load accepted form types from:
    - data_dir/config/form_types.json (if data_dir provided), else
    - packaged secfetch/resources/form_types.json
    """
    if data_dir is None:
        return _load_packaged_form_types().accepted_form_types

    # Do not create data/config by default. Use local file only if user already created it.
    json_path = data_dir / "config" / "form_types.json"
    if not json_path.exists():
        return _load_packaged_form_types().accepted_form_types

    payload = json.loads(json_path.read_text(encoding="utf-8"))
    accepted = payload.get("accepted_form_types")
    if not isinstance(accepted, list) or not all(isinstance(x, str) for x in accepted):
        raise FormTypeValidationError(f"Invalid form types file: {json_path}")
    return sorted(set(x.strip() for x in accepted if x.strip()))


def validate_forms(*, forms: Sequence[str], accepted: Iterable[str]) -> List[str]:
    accepted_set = {x.strip() for x in accepted}
    requested = [f.strip() for f in forms if f and f.strip()]
    if not requested:
        raise FormTypeValidationError("forms must be a non-empty list of SEC form types")

    unknown = sorted({f for f in requested if f not in accepted_set})
    if unknown:
        sample = ", ".join(sorted(list(accepted_set))[:12])
        raise FormTypeValidationError(
            "Unknown/unsupported form types: "
            + ", ".join(unknown)
            + f". Allowed forms come from form_types.json (e.g. {sample}, ...)."
        )
    return requested



from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, Optional


@dataclass(frozen=True)
class ManifestEntry:
    accession: str
    form_type: str
    cik: str
    date_filed: str
    strategy: str


class Manifest:
    def __init__(self, path: Path) -> None:
        self._path = path
        self._data: Dict[str, ManifestEntry] = {}

    @property
    def path(self) -> Path:
        return self._path

    def load(self) -> None:
        if not self._path.exists():
            self._data = {}
            return
        raw = json.loads(self._path.read_text())
        out: Dict[str, ManifestEntry] = {}
        for accession, payload in raw.items():
            out[accession] = ManifestEntry(
                accession=accession,
                form_type=payload["form_type"],
                cik=payload["cik"],
                date_filed=payload["date_filed"],
                strategy=payload.get("strategy", "index"),
            )
        self._data = out

    def has(self, accession: str) -> bool:
        return accession in self._data

    def get(self, accession: str) -> Optional[ManifestEntry]:
        return self._data.get(accession)

    def upsert(self, entry: ManifestEntry) -> None:
        self._data[entry.accession] = entry

    def save_atomic(self) -> None:
        self._path.parent.mkdir(parents=True, exist_ok=True)
        tmp = self._path.with_suffix(self._path.suffix + ".tmp")
        payload = {acc: asdict(ent) for acc, ent in self._data.items()}
        tmp.write_text(json.dumps(payload, indent=2, sort_keys=True))
        tmp.replace(self._path)




from __future__ import annotations

import asyncio
import json
import os
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Optional

import httpx

from secfetch.exceptions import MissingUserAgentError, RateLimitedError
from secfetch.network.rate_limit import RateLimiter


def _user_agent_from_email_json(data_dir: Optional[Path] = None) -> str:
    """Load emails from data_dir/config/email.json if it exists, else package default. Never creates config folder."""
    base = (data_dir if data_dir is not None else Path.cwd() / "data").resolve()
    path = base / "config" / "email.json"
    raw: Optional[str] = None
    if path.is_file():
        try:
            raw = path.read_text(encoding="utf-8")
        except OSError:
            pass
    if raw is None:
        try:
            from importlib.resources import files
            raw = (files("secfetch.resources") / "email.json").read_text(encoding="utf-8")
        except Exception:
            return ""
    if not raw:
        return ""
    try:
        data = json.loads(raw)
        emails: List[str] = data.get("emails") if isinstance(data, dict) else []
        if isinstance(emails, list) and emails:
            valid = [e for e in emails if isinstance(e, str) and "@" in e]
            if valid:
                return f"sec-fetcher {random.choice(valid)}"
    except json.JSONDecodeError:
        pass
    return ""


@dataclass(frozen=True)
class SecClientConfig:
    user_agent: str
    max_requests_per_second: float = 8.0
    timeout_seconds: float = 30.0
    max_retries: int = 6


class SecClient:
    """
    Minimal SEC-safe HTTP client:
    - required User-Agent
    - global rate limiting
    - retry/backoff on 429/5xx
    """

    def __init__(self, config: SecClientConfig):
        if not config.user_agent or "@" not in config.user_agent:
            raise MissingUserAgentError(
                'SEC user-agent is required (include contact info), e.g. "My Org contact@example.com".'
            )

        self._config = config
        self._rate_limiter = RateLimiter(config.max_requests_per_second)
        self._client = httpx.AsyncClient(
            follow_redirects=True,
            timeout=httpx.Timeout(config.timeout_seconds),
            headers={
                "User-Agent": config.user_agent,
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Language": "en-US,en;q=0.9",
                "Connection": "keep-alive",
            },
        )

    @classmethod
    def from_env(
        cls,
        *,
        user_agent: Optional[str] = None,
        data_dir: Optional[Path] = None,
    ) -> "SecClient":
        ua = (
            user_agent
            or os.getenv("SEC_USER_AGENT")
            or _user_agent_from_email_json(data_dir=data_dir)
        )
        return cls(SecClientConfig(user_agent=ua))

    async def aclose(self) -> None:
        await self._client.aclose()

    async def get_bytes(self, url: str) -> bytes:
        resp = await self._request("GET", url)
        return resp.content

    async def get_text(self, url: str) -> str:
        resp = await self._request("GET", url)
        return resp.text

    async def get_json(self, url: str) -> Any:
        resp = await self._request("GET", url)
        return resp.json()

    async def _request(self, method: str, url: str) -> httpx.Response:
        last_exc: Optional[Exception] = None

        for attempt in range(1, self._config.max_retries + 1):
            await self._rate_limiter.wait()
            try:
                resp = await self._client.request(method, url)

                if resp.status_code == 429:
                    # SEC rate limit: obey Retry-After when present, else backoff.
                    retry_after = resp.headers.get("Retry-After")
                    if retry_after and retry_after.isdigit():
                        sleep_s = float(retry_after)
                    else:
                        sleep_s = min(60.0, (2**attempt) + random.random())
                    await asyncio.sleep(sleep_s)
                    last_exc = RateLimitedError(f"429 rate limited for {url}")
                    continue

                if 500 <= resp.status_code < 600:
                    sleep_s = min(30.0, (2**attempt) * 0.5 + random.random())
                    await asyncio.sleep(sleep_s)
                    last_exc = httpx.HTTPStatusError(
                        f"Server error {resp.status_code} for {url}",
                        request=resp.request,
                        response=resp,
                    )
                    continue

                resp.raise_for_status()
                return resp

            except (httpx.TimeoutException, httpx.NetworkError, httpx.HTTPStatusError) as e:
                last_exc = e
                sleep_s = min(10.0, attempt * 0.5 + random.random())
                await asyncio.sleep(sleep_s)
                continue

        if last_exc:
            raise last_exc
        raise RuntimeError(f"Request failed without exception: {method} {url}")




from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import Iterable, List

from secfetch.exceptions import MasterIndexParseError
from secfetch.network.client import SecClient


@dataclass(frozen=True)
class MasterIndexRow:
    cik: str
    company_name: str
    form_type: str
    date_filed: date
    filename: str  # edgar path under /Archives/

    @property
    def accession(self) -> str:
        # filename is typically: edgar/data/{cik}/{accession_no_dash}/{accession}.txt
        name = Path(self.filename).name
        return name.removesuffix(".txt").removesuffix(".idx")

    @property
    def accession_no_dash(self) -> str:
        return self.accession.replace("-", "")


def master_index_url(*, year: int, quarter: int) -> str:
    return f"https://www.sec.gov/Archives/edgar/full-index/{year}/QTR{quarter}/master.idx"


def master_index_cache_path(*, data_dir: Path, year: int, quarter: int) -> Path:
    return data_dir / "index" / "master" / str(year) / f"QTR{quarter}" / "master.idx"


async def download_master_index(
    client: SecClient,
    *,
    data_dir: Path,
    year: int,
    quarter: int,
    force: bool = False,
) -> Path:
    """
    Download and cache master.idx for a given year/quarter.
    Returns the path to the cached file.
    """
    cache_path = master_index_cache_path(data_dir=data_dir, year=year, quarter=quarter)
    if cache_path.exists() and not force:
        return cache_path

    cache_path.parent.mkdir(parents=True, exist_ok=True)
    content = await client.get_bytes(master_index_url(year=year, quarter=quarter))
    cache_path.write_bytes(content)
    return cache_path


def parse_master_index(lines: Iterable[str]) -> List[MasterIndexRow]:
    """
    Parse SEC master.idx into structured rows.
    """
    saw_header = False
    in_data = False
    rows: List[MasterIndexRow] = []

    for raw in lines:
        line = raw.rstrip("\n")
        if not in_data:
            if line.startswith("CIK|Company Name|Form Type|Date Filed|Filename"):
                saw_header = True
                continue
            # Data starts after the delimiter line of dashes that follows the header.
            if saw_header and line.strip().startswith("----"):
                in_data = True
            continue

        if not line.strip():
            continue

        parts = line.split("|")
        if len(parts) != 5:
            raise MasterIndexParseError(f"Unexpected master.idx row format: {line!r}")

        cik, company_name, form_type, date_filed_str, filename = parts
        try:
            yyyy, mm, dd = (int(x) for x in date_filed_str.split("-"))
            dt = date(yyyy, mm, dd)
        except Exception as e:
            raise MasterIndexParseError(
                f"Invalid date in master.idx row: {date_filed_str!r}"
            ) from e

        rows.append(
            MasterIndexRow(
                cik=cik.strip(),
                company_name=company_name.strip(),
                form_type=form_type.strip(),
                date_filed=dt,
                filename=filename.strip(),
            )
        )

    if not rows:
        raise MasterIndexParseError("No rows parsed from master.idx (header not found?)")
    return rows


def load_master_index(path: Path) -> List[MasterIndexRow]:
    text = path.read_text(errors="replace")
    return parse_master_index(text.splitlines())


def iter_unique_accessions(rows: Iterable[MasterIndexRow]) -> Iterable[MasterIndexRow]:
    seen: set[str] = set()
    for row in rows:
        if row.accession in seen:
            continue
        seen.add(row.accession)
        yield row




from __future__ import annotations

import argparse
import json
import sys

from secfetch import download_quarter, download_year
from secfetch.downloader import DownloadResult

# Spinner chars for loading style (cycle per completion)
SPINNER = "|/-\\"

# ANSI colors for progress (no-op if not a TTY)
CYAN = "\033[36m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
MAGENTA = "\033[35m"
BOLD = "\033[1m"
RESET = "\033[0m"


def _progress_callback(
    current: int, total: int, result: DownloadResult | None, in_progress: int = 0
) -> None:
    if total == 0:
        return
    spin = "..." if result is None else SPINNER[(current - 1) % len(SPINNER)]
    # Colourful: label in cyan, count in green, spinner in yellow, in_progress in magenta
    in_progress_part = f" {MAGENTA}({in_progress} downloading){RESET}" if in_progress else ""
    msg = f"\r  {CYAN}{BOLD}Processing{RESET} {GREEN}{current}/{total}{RESET} {YELLOW}{spin}{RESET}{in_progress_part}  "
    sys.stderr.write(msg)
    sys.stderr.flush()


def main(argv: list[str] | None = None) -> int:
    p = argparse.ArgumentParser(prog="secfetch")
    sub = p.add_subparsers(dest="cmd", required=True)

    q = sub.add_parser("quarter", help="Download filings for a quarter")
    q.add_argument("--year", type=int, required=True)
    q.add_argument("--quarter", type=int, required=True, choices=[1, 2, 3, 4])
    q.add_argument("--forms", nargs="+", required=True)
    q.add_argument("--cik", nargs="+", default=None)
    q.add_argument("--ticker", nargs="+", default=None)
    q.add_argument("--data-dir", default="data")
    q.add_argument("--file-types", nargs="+", default=[".htm", ".html", ".xml", ".xbrl", ".pdf"])
    q.add_argument("--include-amended", action="store_true")
    q.add_argument("--concurrency", type=int, default=6)
    q.add_argument("--user-agent", default=None)

    y = sub.add_parser("year", help="Download filings for a year (all quarters)")
    y.add_argument("--year", type=int, required=True)
    y.add_argument("--forms", nargs="+", required=True)
    y.add_argument("--cik", nargs="+", default=None)
    y.add_argument("--ticker", nargs="+", default=None)
    y.add_argument("--data-dir", default="data")
    y.add_argument("--file-types", nargs="+", default=[".htm", ".html", ".xml", ".xbrl", ".pdf"])
    y.add_argument("--include-amended", action="store_true")
    y.add_argument("--concurrency", type=int, default=6)
    y.add_argument("--user-agent", default=None)

    args = p.parse_args(argv)

    try:
        if args.cmd == "quarter":
            res = download_quarter(
                year=args.year,
                quarter=args.quarter,
                forms=args.forms,
                cik=args.cik,
                ticker=args.ticker,
                data_dir=args.data_dir,
                file_types=args.file_types,
                include_amended=args.include_amended,
                concurrency=args.concurrency,
                user_agent=args.user_agent,
                on_progress=_progress_callback,
            )
        else:
            res = download_year(
                year=args.year,
                forms=args.forms,
                cik=args.cik,
                ticker=args.ticker,
                data_dir=args.data_dir,
                file_types=args.file_types,
                include_amended=args.include_amended,
                concurrency=args.concurrency,
                user_agent=args.user_agent,
                on_progress=_progress_callback,
            )
    finally:
        # Clear progress line so JSON output starts clean
        sys.stderr.write("\r" + " " * 60 + "\r")
        sys.stderr.flush()

    print(json.dumps([r.__dict__ for r in res], indent=2, default=str))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import asyncio
import time


class RateLimiter:
    """
    Simple global rate limiter (shared across requests).

    Enforces an average max rate by spacing requests at 1/max_per_second.
    """

    def __init__(self, max_per_second: float) -> None:
        if max_per_second <= 0:
            raise ValueError("max_per_second must be > 0")
        self._interval = 1.0 / max_per_second
        self._lock = asyncio.Lock()
        self._next_time = 0.0

    async def wait(self) -> None:
        async with self._lock:
            now = time.monotonic()
            if now < self._next_time:
                await asyncio.sleep(self._next_time - now)
            now2 = time.monotonic()
            self._next_time = max(self._next_time, now2) + self._interval




from __future__ import annotations

SEC_ARCHIVES_BASE = "https://www.sec.gov/Archives/"


def accession_no_dash(accession: str) -> str:
    return accession.replace("-", "")


def filing_folder_url(*, cik: str, accession: str) -> str:
    # Folder: /Archives/edgar/data/{CIK}/{ACCESSION_NO_DASH}/
    return f"{SEC_ARCHIVES_BASE}edgar/data/{int(cik)}/{accession_no_dash(accession)}/"


def filing_index_json_url(*, cik: str, accession: str) -> str:
    # Folder listing JSON: /Archives/edgar/data/{CIK}/{ACCESSION_NO_DASH}/index.json
    return filing_folder_url(cik=cik, accession=accession) + "index.json"



from __future__ import annotations

import asyncio
import os
import sys
import tarfile
import shutil
from datetime import date
from pathlib import Path
from typing import Callable, List, Optional, Sequence

import httpx

from secfetch.downloader import DownloadResult, FilingDownloader
from secfetch.entities import resolve_cik_filter, resolve_output_group_label
from secfetch.forms import load_accepted_form_types, validate_forms
from secfetch.index.filter import FilingFilter, filter_master_rows
from secfetch.index.master import MasterIndexRow, download_master_index, load_master_index
from secfetch.network.client import SecClient
from secfetch.storage.layout import filing_dir


def _progress_bar(stage: str, current: int, total: int, extra: str = "") -> None:
    total = max(total, 1)
    current = max(0, min(current, total))
    width = 30
    filled = int(width * current / total)
    bar = "=" * filled + "." * (width - filled)
    cyan = "\033[36m"
    magenta = "\033[35m"
    reset = "\033[0m"
    extra_part = f" {extra}" if extra else ""
    sys.stderr.write(f"\r{magenta}: {stage}{reset} [{cyan}{bar}{reset}] {current}/{total}{extra_part}")
    if current >= total:
        sys.stderr.write("\n")
    sys.stderr.flush()


def _step_done(text: str) -> None:
    green = "\033[32m"
    reset = "\033[0m"
    sys.stderr.write(f"{green}[ok]{reset} {text}\n")
    sys.stderr.flush()


def _step_info(text: str) -> None:
    cyan = "\033[36m"
    reset = "\033[0m"
    sys.stderr.write(f"{cyan}[start]{reset} {text}\n")
    sys.stderr.flush()


def _render_filter_label(
    *,
    forms: Sequence[str],
    cik: Optional[str | int | Sequence[str | int]],
    ticker: Optional[str | Sequence[str]],
) -> str:
    form_text = ",".join(forms)
    if ticker is not None:
        vals = [ticker] if isinstance(ticker, str) else list(ticker)
        clean = [str(v).strip().upper() for v in vals if str(v).strip()]
        if clean:
            return f"forms={form_text} ticker={','.join(clean)}"
    if cik is not None:
        vals = [cik] if isinstance(cik, (str, int)) else list(cik)
        clean = [str(v).strip() for v in vals if str(v).strip()]
        if clean:
            return f"forms={form_text} cik={','.join(clean)}"
    return f"forms={form_text}"


def _default_progress_callback(current: int, total: int, result: Optional[DownloadResult], in_progress: int = 0) -> None:
    if total == 0:
        return
    _progress_bar("downloading filings", current, total, "")


async def _run_with_downloader(
    *,
    runner,
    forms: Sequence[str],
    data_dir: str | Path,
    file_types: Sequence[str],
    include_amended: bool,
    cik: Optional[str | int | Sequence[str | int]],
    ticker: Optional[str | Sequence[str]],
    concurrency: int,
    user_agent: Optional[str],
    manifest_path: Optional[str | Path],
    output_format: str = "files",
    on_progress: Optional[Callable[[int, int, Optional[DownloadResult], int], None]] = None,
) -> List[DownloadResult]:
    dl = FilingDownloader(
        forms=forms,
        data_dir=data_dir,
        file_types=file_types,
        include_amended=include_amended,
        cik=cik,
        ticker=ticker,
        concurrency=concurrency,
        user_agent=user_agent,
        manifest_path=manifest_path,
        output_format=output_format,
        on_progress=on_progress,
    )
    try:
        return await runner(dl)
    finally:
        await dl.aclose()


async def _collect_matched_rows_for_quarter(
    *,
    year: int,
    quarter: int,
    forms: Sequence[str],
    data_dir: Path,
    include_amended: bool,
    cik: Optional[str | int | Sequence[str | int]],
    ticker: Optional[str | Sequence[str]],
    user_agent: Optional[str],
) -> tuple[List[MasterIndexRow], Path]:
    accepted = load_accepted_form_types(data_dir=data_dir)
    valid_forms = validate_forms(forms=forms, accepted=accepted)
    client = SecClient.from_env(user_agent=user_agent, data_dir=data_dir)
    try:
        master_path = await download_master_index(client, data_dir=data_dir, year=year, quarter=quarter)
        rows = load_master_index(master_path)
        flt = FilingFilter(forms=valid_forms, include_amended=include_amended)
        matched = filter_master_rows(rows, flt)
        cik_set = resolve_cik_filter(cik=cik, ticker=ticker)
        if cik_set is not None:
            matched = [r for r in matched if r.cik.zfill(10) in cik_set]
        return matched, master_path
    finally:
        await client.aclose()


async def _collect_latest_row_for_company(
    *,
    cik: str,
    user_agent: Optional[str],
    data_dir: Path,
) -> Optional[MasterIndexRow]:
    client = SecClient.from_env(user_agent=user_agent, data_dir=data_dir)
    try:
        cik10 = str(int(str(cik).strip())).zfill(10)
        payload = await client.get_json(f"https://data.sec.gov/submissions/CIK{cik10}.json")
        filings = payload.get("filings", {})
        recent = filings.get("recent", {}) if isinstance(filings, dict) else {}
        acc = recent.get("accessionNumber", [])
        forms = recent.get("form", [])
        dates = recent.get("filingDate", [])
        if not isinstance(acc, list) or not acc:
            return None
        accession = str(acc[0]).strip()
        form = str(forms[0]).strip() if isinstance(forms, list) and forms else ""
        filed = str(dates[0]).strip() if isinstance(dates, list) and dates else ""
        try:
            filed_dt = date.fromisoformat(filed)
        except Exception:
            filed_dt = date.today()
        filename = f"edgar/data/{int(cik10)}/{accession.replace('-', '')}/{accession}.txt"
        return MasterIndexRow(
            cik=cik10,
            company_name=str(payload.get("name", "")).strip(),
            form_type=form or "UNKNOWN",
            date_filed=filed_dt,
            filename=filename,
        )
    finally:
        await client.aclose()


async def _download_datamule_tars_async(
    *,
    rows: Sequence[MasterIndexRow],
    out_dir: Path,
    api_key: Optional[str],
    show_progress: bool,
    concurrency: int,
) -> List[DownloadResult]:
    base_url = "https://sec-library.tar.datamule.xyz/"
    out_dir.mkdir(parents=True, exist_ok=True)
    key = api_key or os.getenv("DATAMULE_API_KEY")
    headers = {"Connection": "keep-alive", "Accept-Encoding": "gzip, deflate, br"}
    if key:
        headers["Authorization"] = f"Bearer {key}"

    total = len(rows)
    sem = asyncio.Semaphore(max(1, int(concurrency)))
    progress_lock = asyncio.Lock()
    done = 0
    downloaded = 0
    skipped = 0
    errors = 0
    active = 0
    started_at = asyncio.get_event_loop().time()

    def _render_progress(current_acc: str) -> None:
        _progress_bar("downloading tar files", done, total, f"{current_acc[:18]}")

    async def one(r: MasterIndexRow) -> DownloadResult:
        nonlocal done, downloaded, skipped, errors, active
        acc_no_dash = r.accession.replace("-", "").zfill(18)
        tar_name = f"{acc_no_dash}.tar"
        tar_path = out_dir / tar_name
        if tar_path.exists():
            async with progress_lock:
                skipped += 1
                done += 1
                if show_progress:
                    _render_progress(r.accession)
            return DownloadResult(
                accession=r.accession,
                cik=r.cik,
                form_type=r.form_type,
                date_filed=r.date_filed,
                status="skipped",
                output_dir=str(tar_path),
            )

        url = base_url + tar_name
        async with sem:
            async with progress_lock:
                active += 1
                if show_progress:
                    _render_progress(r.accession)
            try:
                async with client.stream("GET", url) as resp:
                    resp.raise_for_status()
                    with tar_path.open("wb") as f:
                        async for chunk in resp.aiter_bytes():
                            if chunk:
                                f.write(chunk)
                result = DownloadResult(
                    accession=r.accession,
                    cik=r.cik,
                    form_type=r.form_type,
                    date_filed=r.date_filed,
                    status="downloaded",
                    output_dir=str(tar_path),
                )
                async with progress_lock:
                    downloaded += 1
            except Exception as e:
                try:
                    if tar_path.exists():
                        tar_path.unlink()
                except Exception:
                    pass
                result = DownloadResult(
                    accession=r.accession,
                    cik=r.cik,
                    form_type=r.form_type,
                    date_filed=r.date_filed,
                    status="error",
                    error=str(e),
                    output_dir=str(tar_path),
                )
                async with progress_lock:
                    errors += 1
            finally:
                async with progress_lock:
                    active = max(0, active - 1)
        async with progress_lock:
            done += 1
            if show_progress:
                _render_progress(r.accession)
        return result

    async with httpx.AsyncClient(follow_redirects=True, timeout=120.0, headers=headers) as client:
        results = await asyncio.gather(*[one(r) for r in rows])

    if show_progress and total > 0:
        _progress_bar("downloading tar files", total, total, "")
        _step_done(f"downloaded={downloaded} skipped={skipped} errors={errors}")
    return results


def _safe_extract_tar_to_accession(*, tar_path: Path, target_dir: Path, accession: str) -> None:
    target_dir.mkdir(parents=True, exist_ok=True)
    with tarfile.open(tar_path, mode="r:*") as tf:
        members = tf.getmembers()
        for m in members:
            if not m.isfile():
                continue
            name = m.name.lstrip("/").replace("\\", "/")
            parts = [p for p in name.split("/") if p and p != "."]
            if not parts:
                continue
            if len(parts) > 1:
                head = parts[0]
                head_norm = head.replace("-", "")
                if head_norm.isdigit() and len(head_norm) in (18, 20):
                    parts = parts[1:]
            rel = Path(*parts)
            out_path = (target_dir / rel).resolve()
            if not str(out_path).startswith(str(target_dir.resolve())):
                continue
            out_path.parent.mkdir(parents=True, exist_ok=True)
            src = tf.extractfile(m)
            if src is None:
                continue
            with out_path.open("wb") as dst:
                shutil.copyfileobj(src, dst)


def _extract_and_cleanup_datamule_tars(
    *,
    results: List[DownloadResult],
    data_dir: Path,
    tar_dir: Path,
    show_progress: bool,
    group_label: Optional[str],
) -> List[DownloadResult]:
    updated: List[DownloadResult] = []
    candidates = [r for r in results if r.status != "error" and r.output_dir and Path(r.output_dir).suffix.lower() == ".tar"]
    total = len(candidates)
    done = 0
    for r in results:
        if r.status == "error" or not r.output_dir:
            updated.append(r)
            continue
        tar_path = Path(r.output_dir)
        if not tar_path.exists() or tar_path.suffix.lower() != ".tar":
            updated.append(r)
            continue
        dest = filing_dir(
            data_dir=data_dir,
            form_type=r.form_type,
            cik=r.cik,
            accession=r.accession,
            group_label=group_label,
        )
        try:
            if dest.exists():
                shutil.rmtree(dest)
            _safe_extract_tar_to_accession(tar_path=tar_path, target_dir=dest, accession=r.accession)
            tar_path.unlink(missing_ok=True)
            done += 1
            if show_progress and total > 0:
                _progress_bar("extracting filings", done, total, r.accession)
            updated.append(
                DownloadResult(
                    accession=r.accession,
                    cik=r.cik,
                    form_type=r.form_type,
                    date_filed=r.date_filed,
                    status=r.status,
                    error=r.error,
                    output_dir=str(dest),
                )
            )
        except Exception as e:
            done += 1
            if show_progress and total > 0:
                _progress_bar("extracting filings", done, total, r.accession)
            updated.append(
                DownloadResult(
                    accession=r.accession,
                    cik=r.cik,
                    form_type=r.form_type,
                    date_filed=r.date_filed,
                    status="error",
                    error=str(e),
                    output_dir=str(dest),
                )
            )
    shutil.rmtree(tar_dir, ignore_errors=True)
    return updated


def download_quarter(
    *,
    year: Optional[int] = None,
    quarter: Optional[int] = None,
    forms: Optional[Sequence[str]] = None,
    data_dir: str | Path = "data",
    file_types: Sequence[str] = (".htm", ".html", ".xml", ".xbrl", ".pdf"),
    include_amended: bool = False,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
    concurrency: int = 6,
    user_agent: Optional[str] = None,
    manifest_path: Optional[str | Path] = None,
    show_progress: bool = True,
    on_progress: Optional[Callable[[int, int, Optional[DownloadResult], int], None]] = None,
) -> List[DownloadResult]:
    latest_mode = year is None and quarter is None and forms is None
    if latest_mode and cik is None and ticker is None:
        raise ValueError("For latest single filing mode, provide `cik` or `ticker`.")
    if not latest_mode and (year is None or quarter is None or forms is None):
        raise ValueError("Provide `year`, `quarter`, and `forms` together, or omit all three for latest mode.")

    forms_for_log = forms if forms is not None else ["latest"]
    if show_progress:
        _step_info(
            (
                "latest single download "
                if latest_mode
                else f"quarter download year={year} q={quarter} "
            )
            + _render_filter_label(forms=forms_for_log, cik=cik, ticker=ticker)
        )

    progress_cb = on_progress if on_progress is not None else (_default_progress_callback if show_progress else None)

    if latest_mode:
        data_dir_path = Path(data_dir)
        cik_set = resolve_cik_filter(cik=cik, ticker=ticker)
        chosen_cik = sorted(cik_set)[0]
        latest_row = asyncio.run(
            _collect_latest_row_for_company(
                cik=chosen_cik,
                user_agent=user_agent,
                data_dir=data_dir_path,
            )
        )
        if latest_row is None:
            return []

        async def _run_latest() -> List[DownloadResult]:
            return await _run_with_downloader(
                runner=lambda dl: dl._download_one(row=latest_row),
                forms=[latest_row.form_type or "8-K"],
                data_dir=data_dir,
                file_types=file_types,
                include_amended=include_amended,
                cik=cik,
                ticker=ticker,
                concurrency=concurrency,
                user_agent=user_agent,
                manifest_path=manifest_path,
                on_progress=progress_cb,
            )

        return asyncio.run(_run_latest())

    async def _run() -> List[DownloadResult]:
        return await _run_with_downloader(
            runner=lambda dl: dl.download_quarter(year=int(year), quarter=int(quarter)),
            forms=forms,
            data_dir=data_dir,
            file_types=file_types,
            include_amended=include_amended,
            cik=cik,
            ticker=ticker,
            concurrency=concurrency,
            user_agent=user_agent,
            manifest_path=manifest_path,
            on_progress=progress_cb,
        )

    return asyncio.run(_run())


def download_quarter_tar(
    *,
    year: Optional[int] = None,
    quarter: Optional[int] = None,
    forms: Optional[Sequence[str]] = None,
    data_dir: str | Path = "data",
    file_types: Sequence[str] = (".htm", ".html", ".xml", ".xbrl", ".pdf"),
    include_amended: bool = False,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
    concurrency: int = 20,
    user_agent: Optional[str] = None,
    manifest_path: Optional[str | Path] = None,
    output_dir: Optional[str | Path] = None,
    datamule_api_key: Optional[str] = None,
    tar_provider: str = "datamule",
    limit: Optional[int] = None,
    extract: bool = True,
    show_progress: bool = True,
    on_progress: Optional[Callable[[int, int, Optional[DownloadResult], int], None]] = None,
) -> List[DownloadResult]:
    latest_mode = year is None and quarter is None and forms is None
    if latest_mode and cik is None and ticker is None:
        raise ValueError("For latest single filing mode, provide `cik` or `ticker`.")
    if not latest_mode and (year is None or quarter is None or forms is None):
        raise ValueError("Provide `year`, `quarter`, and `forms` together, or omit all three for latest mode.")

    forms_for_log = forms if forms is not None else ["latest"]
    if show_progress:
        _step_info(
            (
                f"latest single tar download "
                if latest_mode
                else f"quarter tar download year={year} q={quarter} "
            )
            + _render_filter_label(forms=forms_for_log, cik=cik, ticker=ticker)
        )
    if tar_provider == "local":
        if latest_mode:
            raise ValueError("latest-mode without year/quarter/forms is only supported with tar_provider='datamule'")
        if forms is None or year is None or quarter is None:
            raise ValueError("year, quarter, and forms are required for tar_provider='local'")
        progress_cb = on_progress if on_progress is not None else (_default_progress_callback if show_progress else None)

        async def _run_local() -> List[DownloadResult]:
            return await _run_with_downloader(
                runner=lambda dl: dl.download_quarter(year=int(year), quarter=int(quarter)),
                forms=forms,
                data_dir=data_dir,
                file_types=file_types,
                include_amended=include_amended,
                cik=cik,
                ticker=ticker,
                concurrency=concurrency,
                user_agent=user_agent,
                manifest_path=manifest_path,
                output_format="tar",
                on_progress=progress_cb,
            )

        return asyncio.run(_run_local())

    if tar_provider != "datamule":
        raise ValueError("tar_provider must be 'datamule' or 'local'")

    data_dir_path = Path(data_dir)
    group_label = resolve_output_group_label(cik=cik, ticker=ticker)
    out_dir = Path(output_dir) if output_dir is not None else (data_dir_path / "filings_tar")

    if latest_mode:
        cik_set = resolve_cik_filter(cik=cik, ticker=ticker)
        chosen_cik = sorted(cik_set)[0]
        latest_row = asyncio.run(
            _collect_latest_row_for_company(
                cik=chosen_cik,
                user_agent=user_agent,
                data_dir=data_dir_path,
            )
        )
        rows = [latest_row] if latest_row is not None else []
        master_path = None
    else:
        rows, master_path = asyncio.run(
            _collect_matched_rows_for_quarter(
                year=int(year),
                quarter=int(quarter),
                forms=forms,
                data_dir=data_dir_path,
                include_amended=include_amended,
                cik=cik,
                ticker=ticker,
                user_agent=user_agent,
            )
        )
    if not rows:
        return []
    if show_progress:
        _step_done(f"queued {len(rows)} filings")
    if limit is not None:
        limit = int(limit)
        if limit <= 0:
            return []
        rows = rows[:limit]

    accessions = [r.accession for r in rows]
    # Keep order and de-duplicate.
    accessions = list(dict.fromkeys(accessions))

    # Datamule mode: fetch source tar files directly from DataMule tar endpoint.
    # `file_types` filtering is not applied in this mode.
    results = asyncio.run(
        _download_datamule_tars_async(
            rows=rows,
            out_dir=out_dir,
            api_key=datamule_api_key,
            show_progress=show_progress,
            concurrency=concurrency,
        )
    )
    if extract:
        results = _extract_and_cleanup_datamule_tars(
            results=results,
            data_dir=data_dir_path,
            tar_dir=out_dir,
            show_progress=show_progress,
            group_label=group_label,
        )

    # Match existing quarter behavior: clear quarter index cache when run completes.
    if master_path is not None:
        try:
            if master_path.parent.exists():
                import shutil

                shutil.rmtree(master_path.parent)
        except Exception:
            pass

    return results


def download_year(
    *,
    year: int,
    forms: Sequence[str],
    data_dir: str | Path = "data",
    file_types: Sequence[str] = (".htm", ".html", ".xml", ".xbrl", ".pdf"),
    include_amended: bool = False,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
    concurrency: int = 6,
    user_agent: Optional[str] = None,
    manifest_path: Optional[str | Path] = None,
    quarters: Sequence[int] = (1, 2, 3, 4),
    show_progress: bool = True,
    on_progress: Optional[Callable[[int, int, Optional[DownloadResult], int], None]] = None,
) -> List[DownloadResult]:
    if show_progress:
        _step_info(
            f"year download year={year} "
            + _render_filter_label(forms=forms, cik=cik, ticker=ticker)
        )
    progress_cb = on_progress if on_progress is not None else (_default_progress_callback if show_progress else None)

    async def _run() -> List[DownloadResult]:
        return await _run_with_downloader(
            runner=lambda dl: dl.download_year(year=year, quarters=quarters),
            forms=forms,
            data_dir=data_dir,
            file_types=file_types,
            include_amended=include_amended,
            cik=cik,
            ticker=ticker,
            concurrency=concurrency,
            user_agent=user_agent,
            manifest_path=manifest_path,
            on_progress=progress_cb,
        )

    return asyncio.run(_run())


def download_year_tar(
    *,
    year: int,
    forms: Sequence[str],
    data_dir: str | Path = "data",
    file_types: Sequence[str] = (".htm", ".html", ".xml", ".xbrl", ".pdf"),
    include_amended: bool = False,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
    concurrency: int = 20,
    user_agent: Optional[str] = None,
    manifest_path: Optional[str | Path] = None,
    output_dir: Optional[str | Path] = None,
    datamule_api_key: Optional[str] = None,
    tar_provider: str = "datamule",
    limit: Optional[int] = None,
    quarters: Sequence[int] = (1, 2, 3, 4),
    extract: bool = True,
    show_progress: bool = True,
    on_progress: Optional[Callable[[int, int, Optional[DownloadResult], int], None]] = None,
) -> List[DownloadResult]:
    if show_progress:
        _step_info(
            f"year tar download year={year} "
            + _render_filter_label(forms=forms, cik=cik, ticker=ticker)
        )
    if tar_provider == "local":
        progress_cb = on_progress if on_progress is not None else (_default_progress_callback if show_progress else None)

        async def _run_local() -> List[DownloadResult]:
            return await _run_with_downloader(
                runner=lambda dl: dl.download_year(year=year, quarters=quarters),
                forms=forms,
                data_dir=data_dir,
                file_types=file_types,
                include_amended=include_amended,
                cik=cik,
                ticker=ticker,
                concurrency=concurrency,
                user_agent=user_agent,
                manifest_path=manifest_path,
                output_format="tar",
                on_progress=progress_cb,
            )

        return asyncio.run(_run_local())

    if tar_provider != "datamule":
        raise ValueError("tar_provider must be 'datamule' or 'local'")

    out: List[DownloadResult] = []
    for q in quarters:
        remaining = None if limit is None else max(0, int(limit) - len(out))
        if remaining == 0:
            break
        out.extend(
            download_quarter_tar(
                year=year,
                quarter=int(q),
                forms=forms,
                data_dir=data_dir,
                file_types=file_types,
                include_amended=include_amended,
                cik=cik,
                ticker=ticker,
                concurrency=concurrency,
                user_agent=user_agent,
                manifest_path=manifest_path,
                output_dir=output_dir,
                datamule_api_key=datamule_api_key,
                tar_provider="datamule",
                limit=remaining,
                extract=extract,
                show_progress=show_progress,
                on_progress=on_progress,
            )
        )
    return out



"""
Compatibility module for the PyPI distribution name.

The PyPI package is named `secfetcher`, while the primary Python module is `secfetch`.
This compatibility module intentionally exposes only tar quarter download:

    from secfetcher import download_quarter_tar
"""

from secfetch import download_quarter_tar

__all__ = [
    "download_quarter_tar",
]



from __future__ import annotations

import json
import tarfile
from pathlib import Path

import pytest


def _write_accepted_forms(data_dir: Path, forms: list[str]) -> None:
    p = data_dir / "config" / "form_types.json"
    p.parent.mkdir(parents=True, exist_ok=True)
    p.write_text(json.dumps({"accepted_form_types": forms}, indent=2) + "\n", encoding="utf-8")


def _master_idx_one_row(*, cik: str, form_type: str, date_filed: str, accession: str) -> str:
    # accession must have dashes for SEC format (e.g. 0001000045-24-000001)
    accession_no_dash = accession.replace("-", "")
    return "\n".join(
        [
            "Description: Master Index of EDGAR Dissemination Feed",
            "Last Data Received: 2024-01-02",
            "CIK|Company Name|Form Type|Date Filed|Filename",
            "--------------------------------------------------------------------------------",
            f"{cik}|TEST CORP|{form_type}|{date_filed}|edgar/data/{int(cik)}/{accession_no_dash}/{accession}.txt",
            "",
        ]
    )


class DummyClient:
    def __init__(self, *, master_idx_text: str, listing: dict, file_bytes: dict[str, bytes]):
        self._master_idx_text = master_idx_text
        self._listing = listing
        self._file_bytes = file_bytes

    async def aclose(self) -> None:
        return None

    async def get_bytes(self, url: str) -> bytes:
        if url.endswith("/master.idx") or url.endswith("master.idx"):
            return self._master_idx_text.encode("utf-8")
        return self._file_bytes[url]

    async def get_json(self, url: str):
        # Only used for /index.json
        assert url.endswith("index.json")
        return self._listing


def test_quarter_download_deletes_index_on_success(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    cik = "1000045"
    accession = "0001000045-24-000001"

    master_idx = _master_idx_one_row(cik=cik, form_type="10-Q", date_filed="2024-01-02", accession=accession)
    listing = {"directory": {"item": [{"name": "doc.xml"}, {"name": "readme.txt"}]}}

    base_folder = f"https://www.sec.gov/Archives/edgar/data/{int(cik)}/{accession.replace('-', '')}/"
    file_bytes = {base_folder + "doc.xml": b"<xml/>", base_folder + "readme.txt": b"nope"}

    from secfetch import download_quarter
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        concurrency=2,
    )

    assert len(res) == 1
    assert res[0].status == "downloaded"

    # Index cache for the quarter should be deleted on success
    qtr_index_dir = data_dir / "index" / "master" / "2024" / "QTR1"
    assert not qtr_index_dir.exists()

    # Downloaded file should exist
    out_file = data_dir / "filings" / "10-Q" / "0001000045" / accession / "doc.xml"
    assert out_file.exists()


def test_quarter_download_keeps_index_on_error(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    cik = "1000045"
    accession = "0001000045-24-000001"
    master_idx = _master_idx_one_row(cik=cik, form_type="10-Q", date_filed="2024-01-02", accession=accession)

    # Listing contains no files matching ".xml" => triggers error
    listing = {"directory": {"item": [{"name": "readme.txt"}]}}
    base_folder = f"https://www.sec.gov/Archives/edgar/data/{int(cik)}/{accession.replace('-', '')}/"
    file_bytes = {base_folder + "readme.txt": b"nope"}

    from secfetch import download_quarter
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        concurrency=1,
    )

    assert len(res) == 1
    assert res[0].status == "error"

    # Index cache should remain when there were errors
    qtr_index_dir = data_dir / "index" / "master" / "2024" / "QTR1"
    assert qtr_index_dir.exists()


def test_quarter_download_filters_by_cik(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    cik_a = "1000045"
    cik_b = "1000046"
    accession_a = "0001000045-24-000001"
    accession_b = "0001000046-24-000001"

    # Two rows in one master index; only cik_a should be downloaded.
    row_a = _master_idx_one_row(cik=cik_a, form_type="10-Q", date_filed="2024-01-02", accession=accession_a)
    row_b = _master_idx_one_row(cik=cik_b, form_type="10-Q", date_filed="2024-01-02", accession=accession_b)
    header, data_a = row_a.split("--------------------------------------------------------------------------------\n", 1)
    _, data_b = row_b.split("--------------------------------------------------------------------------------\n", 1)
    master_idx = header + "--------------------------------------------------------------------------------\n" + data_a.strip() + "\n" + data_b.strip() + "\n"

    listing = {"directory": {"item": [{"name": "doc.xml"}]}}
    base_a = f"https://www.sec.gov/Archives/edgar/data/{int(cik_a)}/{accession_a.replace('-', '')}/"
    base_b = f"https://www.sec.gov/Archives/edgar/data/{int(cik_b)}/{accession_b.replace('-', '')}/"
    file_bytes = {base_a + "doc.xml": b"<xml/>", base_b + "doc.xml": b"<xml/>"}

    from secfetch import download_quarter
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        cik=cik_a,
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        concurrency=2,
        show_progress=False,
    )

    assert len(res) == 1
    assert res[0].cik == cik_a


def test_quarter_download_filters_by_ticker(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    # From listed_filer_metadata.csv: ABT -> CIK 0000001800
    cik_match = "1800"
    cik_other = "1000046"
    accession_match = "0000001800-24-000001"
    accession_other = "0001000046-24-000001"

    row_a = _master_idx_one_row(cik=cik_match, form_type="10-Q", date_filed="2024-01-02", accession=accession_match)
    row_b = _master_idx_one_row(cik=cik_other, form_type="10-Q", date_filed="2024-01-02", accession=accession_other)
    header, data_a = row_a.split("--------------------------------------------------------------------------------\n", 1)
    _, data_b = row_b.split("--------------------------------------------------------------------------------\n", 1)
    master_idx = header + "--------------------------------------------------------------------------------\n" + data_a.strip() + "\n" + data_b.strip() + "\n"

    listing = {"directory": {"item": [{"name": "doc.xml"}]}}
    base_a = f"https://www.sec.gov/Archives/edgar/data/{int(cik_match)}/{accession_match.replace('-', '')}/"
    base_b = f"https://www.sec.gov/Archives/edgar/data/{int(cik_other)}/{accession_other.replace('-', '')}/"
    file_bytes = {base_a + "doc.xml": b"<xml/>", base_b + "doc.xml": b"<xml/>"}

    from secfetch import download_quarter
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        ticker="ABT",
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        show_progress=False,
    )

    assert len(res) == 1
    assert res[0].cik == cik_match


def test_quarter_download_tar_writes_tar_file(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    cik = "1000045"
    accession = "0001000045-24-000001"

    master_idx = _master_idx_one_row(cik=cik, form_type="10-Q", date_filed="2024-01-02", accession=accession)
    listing = {"directory": {"item": [{"name": "doc.xml"}, {"name": "readme.txt"}]}}

    base_folder = f"https://www.sec.gov/Archives/edgar/data/{int(cik)}/{accession.replace('-', '')}/"
    file_bytes = {base_folder + "doc.xml": b"<xml/>", base_folder + "readme.txt": b"nope"}

    from secfetch import download_quarter_tar
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter_tar(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        concurrency=2,
        tar_provider="local",
    )

    assert len(res) == 1
    assert res[0].status == "downloaded"
    tar_path = Path(res[0].output_dir or "")
    assert tar_path.exists()
    assert tar_path.suffix == ".tar"

    with tarfile.open(tar_path, mode="r") as tf:
        names = tf.getnames()
        assert "metadata.json" in names
        assert "doc.xml" in names
        assert "readme.txt" not in names


def test_quarter_download_tar_ignores_old_file_mode_manifest(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    data_dir = tmp_path / "data"
    _write_accepted_forms(data_dir, ["10-Q"])

    cik = "1000045"
    accession = "0001000045-24-000001"
    master_idx = _master_idx_one_row(cik=cik, form_type="10-Q", date_filed="2024-01-02", accession=accession)
    listing = {"directory": {"item": [{"name": "doc.xml"}]}}
    base_folder = f"https://www.sec.gov/Archives/edgar/data/{int(cik)}/{accession.replace('-', '')}/"
    file_bytes = {base_folder + "doc.xml": b"<xml/>"}

    # Simulate previous non-tar run by pre-seeding manifest with strategy=index.
    manifest_path = data_dir / "_state" / "manifest.json"
    manifest_path.parent.mkdir(parents=True, exist_ok=True)
    manifest_path.write_text(
        json.dumps(
            {
                accession: {
                    "accession": accession,
                    "form_type": "10-Q",
                    "cik": "0001000045",
                    "date_filed": "2024-01-02",
                    "strategy": "index",
                }
            }
        ),
        encoding="utf-8",
    )

    from secfetch import download_quarter_tar
    import secfetch.downloader as dl_mod

    monkeypatch.setattr(
        dl_mod.SecClient,
        "from_env",
        classmethod(lambda cls, *, user_agent=None, data_dir=None: DummyClient(master_idx_text=master_idx, listing=listing, file_bytes=file_bytes)),
    )

    res = download_quarter_tar(
        year=2024,
        quarter=1,
        forms=["10-Q"],
        data_dir=data_dir,
        file_types=[".xml"],
        user_agent="Test test@example.com",
        tar_provider="local",
        show_progress=False,
    )

    assert len(res) == 1
    assert res[0].status == "downloaded"
    tar_path = Path(res[0].output_dir or "")
    assert tar_path.exists()



from secfetch.api import download_quarter, download_quarter_tar, download_year, download_year_tar
from secfetch.downloader import FilingDownloader

__all__ = [
    "FilingDownloader",
    "download_quarter",
    "download_quarter_tar",
    "download_year",
    "download_year_tar",
]



from __future__ import annotations

import re
from pathlib import Path


def form_dir_name(form_type: str) -> str:
    """
    Convert a form type into a safe directory name.
    Example: "10-Q/A" -> "10-Q_A"
    """
    s = form_type.strip()
    s = s.replace("/", "_")
    s = re.sub(r"\s+", "", s)
    return s


def filings_root(data_dir: Path) -> Path:
    return data_dir / "filings"


def filing_dir(
    *,
    data_dir: Path,
    form_type: str,
    cik: str,
    accession: str,
    group_label: str | None = None,
) -> Path:
    base = filings_root(data_dir) / form_dir_name(form_type)
    if group_label:
        return base / group_label / accession
    return base / cik.zfill(10) / accession



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, List, Sequence

from secfetch.index.master import MasterIndexRow, iter_unique_accessions


@dataclass(frozen=True)
class FilingFilter:
    forms: Sequence[str]
    include_amended: bool = False

    def match(self, row: MasterIndexRow) -> bool:
        if row.form_type not in self.forms:
            return False
        if not self.include_amended and "/A" in row.form_type:
            return False
        return True


def filter_master_rows(rows: Iterable[MasterIndexRow], flt: FilingFilter) -> List[MasterIndexRow]:
    return [r for r in iter_unique_accessions(rows) if flt.match(r)]




from __future__ import annotations

import asyncio
import io
import json
import shutil
import tarfile
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import Callable, List, Optional, Sequence

from secfetch.entities import resolve_cik_filter, resolve_output_group_label
from secfetch.edgar import filing_index_json_url, filing_folder_url
from secfetch.exceptions import SecFetchError
from secfetch.forms import load_accepted_form_types, validate_forms
from secfetch.index.filter import FilingFilter, filter_master_rows
from secfetch.index.master import (
    MasterIndexRow,
    download_master_index,
    load_master_index,
)
from secfetch.network.client import SecClient
from secfetch.storage.layout import filing_dir
from secfetch.storage.manifest import Manifest, ManifestEntry


class DownloadError(SecFetchError):
    pass


@dataclass(frozen=True)
class DownloadResult:
    accession: str
    cik: str
    form_type: str
    date_filed: date
    status: str  # downloaded | skipped | error
    error: Optional[str] = None
    output_dir: Optional[str] = None


def _normalize_file_types(file_types: Sequence[str]) -> List[str]:
    out: List[str] = []
    for t in file_types:
        tt = t.strip().lower()
        if not tt:
            continue
        if not tt.startswith("."):
            tt = "." + tt
        out.append(tt)
    if not out:
        raise ValueError("file_types must be non-empty (e.g. ['.xml', '.htm', '.html'])")
    return sorted(set(out))


def _default_manifest_path(data_dir: Path) -> Path:
    # keep clean: state under data/_state/
    return data_dir / "_state" / "manifest.json"


def _filing_tar_path(*, data_dir: Path, form_type: str, cik: str, accession: str) -> Path:
    return data_dir / "filings_tar" / form_type / cik.zfill(10) / f"{accession}.tar"


class FilingDownloader:
    """
    Index-driven downloader:
      - downloads master.idx for a quarter
      - filters by form types
      - for each accession: uses EDGAR folder `index.json` to list files
      - downloads only requested file types
      - writes a manifest for de-dup/resume
    """

    def __init__(
        self,
        *,
        forms: Sequence[str],
        data_dir: str | Path = "data",
        file_types: Sequence[str] = (".htm", ".html", ".xml", ".xbrl", ".pdf"),
        include_amended: bool = False,
        cik: Optional[str | int | Sequence[str | int]] = None,
        ticker: Optional[str | Sequence[str]] = None,
        concurrency: int = 6,
        user_agent: Optional[str] = None,
        manifest_path: Optional[str | Path] = None,
        output_format: str = "files",
        on_progress: Optional[Callable[[int, int, Optional["DownloadResult"], int], None]] = None,
    ) -> None:
        self.data_dir = Path(data_dir)
        self.file_types = _normalize_file_types(file_types)
        self.include_amended = include_amended
        self.cik_set = resolve_cik_filter(cik=cik, ticker=ticker)
        self.output_group_label = resolve_output_group_label(cik=cik, ticker=ticker)
        self.concurrency = max(1, int(concurrency))
        self._on_progress = on_progress
        if output_format not in ("files", "tar"):
            raise ValueError("output_format must be 'files' or 'tar'")
        self.output_format = output_format

        accepted = load_accepted_form_types(data_dir=self.data_dir)
        self.forms = validate_forms(forms=forms, accepted=accepted)

        self._client = SecClient.from_env(user_agent=user_agent, data_dir=self.data_dir)
        self._manifest = Manifest(Path(manifest_path) if manifest_path else _default_manifest_path(self.data_dir))
        self._manifest.load()

    async def aclose(self) -> None:
        await self._client.aclose()

    async def download_quarter(self, *, year: int, quarter: int) -> List[DownloadResult]:
        # Always fetch master index first (source of truth)
        master_path = await download_master_index(self._client, data_dir=self.data_dir, year=year, quarter=quarter)
        rows = load_master_index(master_path)

        flt = FilingFilter(forms=self.forms, include_amended=self.include_amended)
        matched = filter_master_rows(rows, flt)
        if self.cik_set is not None:
            matched = [r for r in matched if r.cik.zfill(10) in self.cik_set]
        total = len(matched)

        if self._on_progress is not None and total > 0:
            self._on_progress(0, total, None, 0)

        sem = asyncio.Semaphore(self.concurrency)
        completed = 0
        in_progress = 0
        progress_lock = asyncio.Lock()

        async def with_progress(row: MasterIndexRow):
            nonlocal completed, in_progress
            async with sem:
                async with progress_lock:
                    in_progress += 1
                    if self._on_progress is not None:
                        self._on_progress(completed, total, None, in_progress)
                try:
                    result = await self._download_one(row=row)
                finally:
                    async with progress_lock:
                        in_progress -= 1
                        completed += 1
                        if self._on_progress is not None:
                            self._on_progress(completed, total, result, in_progress)
            return result

        tasks = [with_progress(row) for row in matched]
        results = await asyncio.gather(*tasks)
        self._manifest.save_atomic()

        # If the whole quarter completed without errors, remove cached index data for that quarter.
        # (User requested: delete index data after all files download for a quarter.)
        if all(r.status != "error" for r in results):
            try:
                if master_path.parent.exists():
                    shutil.rmtree(master_path.parent)
            except Exception:
                # Best-effort cleanup; do not fail downloads due to cleanup.
                pass
        return results

    async def download_year(self, *, year: int, quarters: Sequence[int] = (1, 2, 3, 4)) -> List[DownloadResult]:
        out: List[DownloadResult] = []
        for q in quarters:
            out.extend(await self.download_quarter(year=year, quarter=int(q)))
        return out

    async def _download_one(self, *, row: MasterIndexRow) -> DownloadResult:
        accession = row.accession
        out_dir = filing_dir(
            data_dir=self.data_dir,
            form_type=row.form_type,
            cik=row.cik,
            accession=accession,
            group_label=self.output_group_label,
        )
        tar_path = _filing_tar_path(
            data_dir=self.data_dir,
            form_type=row.form_type,
            cik=row.cik,
            accession=accession,
        )
        entry = self._manifest.get(accession)
        if entry is not None:
            if self.output_format == "files":
                if entry.strategy == "index" and out_dir.exists():
                    return DownloadResult(
                        accession=accession,
                        cik=row.cik,
                        form_type=row.form_type,
                        date_filed=row.date_filed,
                        status="skipped",
                    )
            else:
                if entry.strategy == "index_tar" and tar_path.exists():
                    return DownloadResult(
                        accession=accession,
                        cik=row.cik,
                        form_type=row.form_type,
                        date_filed=row.date_filed,
                        status="skipped",
                        output_dir=str(tar_path),
                    )

        tmp_dir = out_dir.with_name(out_dir.name + ".tmp")
        tmp_tar = tar_path.with_suffix(".tmp")
        try:
            if tmp_dir.exists():
                shutil.rmtree(tmp_dir)
            if tmp_tar.exists():
                tmp_tar.unlink()
            tmp_dir.mkdir(parents=True, exist_ok=True)

            index_url = filing_index_json_url(cik=row.cik, accession=accession)
            listing = await self._client.get_json(index_url)
            files = _extract_files_from_index_json(listing, base_folder_url=filing_folder_url(cik=row.cik, accession=accession))

            selected = [f for f in files if _match_file_types(f["name"], self.file_types)]
            if not selected:
                raise DownloadError(
                    f"No files matched file_types={self.file_types} for accession {accession}"
                )

            for f in selected:
                content = await self._client.get_bytes(f["href"])
                (tmp_dir / f["name"]).write_bytes(content)

            if self.output_format == "files":
                # Commit atomically-ish: replace whole folder only after success.
                out_dir.parent.mkdir(parents=True, exist_ok=True)
                if out_dir.exists():
                    shutil.rmtree(out_dir)
                tmp_dir.replace(out_dir)
                output_path = str(out_dir)
                strategy = "index"
            else:
                tmp_tar.parent.mkdir(parents=True, exist_ok=True)
                with tarfile.open(tmp_tar, mode="w") as tf:
                    metadata = {
                        "accession": accession,
                        "cik": row.cik.zfill(10),
                        "form_type": row.form_type,
                        "date_filed": row.date_filed.isoformat(),
                        "files": [f["name"] for f in selected],
                    }
                    metadata_bytes = json.dumps(metadata).encode("utf-8")
                    meta_info = tarfile.TarInfo(name="metadata.json")
                    meta_info.size = len(metadata_bytes)
                    tf.addfile(meta_info, io.BytesIO(metadata_bytes))
                    for f in selected:
                        content = (tmp_dir / f["name"]).read_bytes()
                        info = tarfile.TarInfo(name=f["name"])
                        info.size = len(content)
                        tf.addfile(info, io.BytesIO(content))

                if tar_path.exists():
                    tar_path.unlink()
                tmp_tar.replace(tar_path)
                shutil.rmtree(tmp_dir)
                output_path = str(tar_path)
                strategy = "index_tar"

            self._manifest.upsert(
                ManifestEntry(
                    accession=accession,
                    form_type=row.form_type,
                    cik=row.cik.zfill(10),
                    date_filed=row.date_filed.isoformat(),
                    strategy=strategy,
                )
            )

            return DownloadResult(
                accession=accession,
                cik=row.cik,
                form_type=row.form_type,
                date_filed=row.date_filed,
                status="downloaded",
                output_dir=output_path,
            )
        except Exception as e:
            # Cleanup partials
            try:
                if tmp_dir.exists():
                    shutil.rmtree(tmp_dir)
                if tmp_tar.exists():
                    tmp_tar.unlink()
            except Exception:
                pass
            return DownloadResult(
                accession=accession,
                cik=row.cik,
                form_type=row.form_type,
                date_filed=row.date_filed,
                status="error",
                error=str(e),
            )


def _match_file_types(name: str, file_types: Sequence[str]) -> bool:
    n = name.lower()
    return any(n.endswith(ext) for ext in file_types)


def _extract_files_from_index_json(payload: object, *, base_folder_url: str) -> List[dict]:
    """
    Parse SEC folder index.json response and return:
    [{ "name": "...", "href": "https://..." }, ...]
    """
    if not isinstance(payload, dict):
        raise DownloadError("index.json payload was not an object")

    directory = payload.get("directory")
    if not isinstance(directory, dict):
        raise DownloadError("index.json missing 'directory' object")

    items = directory.get("item")
    if not isinstance(items, list):
        raise DownloadError("index.json missing 'directory.item' array")

    out: List[dict] = []
    for it in items:
        if not isinstance(it, dict):
            continue
        name = it.get("name")
        if not isinstance(name, str) or not name:
            continue
        out.append({"name": name, "href": base_folder_url + name})
    return out



from __future__ import annotations

import ast
import csv
from pathlib import Path
from typing import Dict, Optional, Sequence, Set


def _normalize_cik(value: str | int) -> str:
    s = str(value).strip()
    if not s:
        return ""
    if s.isdigit():
        return str(int(s)).zfill(10)
    return s.zfill(10)


def _load_packaged_listed_filers() -> list[dict[str, str]]:
    from importlib.resources import files

    path = files("secfetch.resources") / "listed_filer_metadata.csv"
    rows: list[dict[str, str]] = []
    with path.open("r", encoding="utf-8", newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            rows.append(row)
    return rows


def load_ticker_to_cik_map() -> Dict[str, Set[str]]:
    out: Dict[str, Set[str]] = {}
    for row in _load_packaged_listed_filers():
        cik = _normalize_cik(row.get("cik", ""))
        if not cik:
            continue
        raw = row.get("tickers", "")
        tickers: list[str] = []
        try:
            parsed = ast.literal_eval(raw) if raw else []
            if isinstance(parsed, list):
                tickers = [str(x).strip().upper() for x in parsed if str(x).strip()]
        except Exception:
            tickers = []
        for t in tickers:
            out.setdefault(t, set()).add(cik)
    return out


def resolve_cik_filter(
    *,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
) -> Optional[Set[str]]:
    values: set[str] = set()
    if cik is not None:
        cik_values = [cik] if isinstance(cik, (str, int)) else list(cik)
        for v in cik_values:
            norm = _normalize_cik(v)
            if norm:
                values.add(norm)

    if ticker is not None:
        ticker_values = [ticker] if isinstance(ticker, str) else list(ticker)
        mapping = load_ticker_to_cik_map()
        for t in ticker_values:
            key = str(t).strip().upper()
            if not key:
                continue
            values.update(mapping.get(key, set()))

    return values or None


def resolve_output_group_label(
    *,
    cik: Optional[str | int | Sequence[str | int]] = None,
    ticker: Optional[str | Sequence[str]] = None,
) -> Optional[str]:
    if ticker is not None:
        tickers = [ticker] if isinstance(ticker, str) else list(ticker)
        cleaned = [str(t).strip().upper() for t in tickers if str(t).strip()]
        if len(cleaned) == 1:
            return cleaned[0]
    if cik is not None:
        ciks = [cik] if isinstance(cik, (str, int)) else list(cik)
        cleaned = [_normalize_cik(v) for v in ciks if str(v).strip()]
        if len(cleaned) == 1:
            return cleaned[0]
    return None



class SecFetchError(Exception):
    """Base error for secfetch."""


class MissingUserAgentError(SecFetchError):
    """Raised when SEC user-agent is not provided."""


class RateLimitedError(SecFetchError):
    """Raised when SEC rate limiting is encountered."""


class MasterIndexParseError(SecFetchError):
    """Raised when master.idx cannot be parsed."""

